import numpy as np
import sys
import os
import logging
import random
from time import time
import scipy.sparse as sp
from scipy.fft import fftn as FFT

from kaldo.forceconstants import ForceConstants
from kaldo.phonons import Phonons
from kaldo.observables.harmonic_with_q import HarmonicWithQ
from kaldo.helpers.storage import get_folder_from_label
from kaldo.grid import Grid
from ase.io import write
import ase.units as units
from ase.atoms import Atoms

logging.basicConfig(level=logging.INFO)


def find_index_from_grid(grid, nrep, grid_type="F"):
    """
    Return a flattened index for grid differences.

    Parameters:
        grid (np.array): Input grid differences.
        nrep (int): Number for modulo operation.
        grid_type (str): 'F' for Fortran-like, 'C' for C-like ordering.

    Returns:
        np.array: Flattened index mask.
    """
    grid = np.mod(grid, nrep)
    if grid_type == "F":
        mask = grid * np.array([1, nrep, nrep**2])
        mask = np.sum(mask, axis=-1)
    if grid_type == "C":
        mask = grid * np.array([nrep**2, nrep, 1])
        mask = np.sum(mask, axis=-1)
    return mask


def from_crystal2real_space_fast(dynmat_crystal, grid, nrep, grid_type="F"):
    """
    Convert dynamical matrix from crystal to real space using fast FFT-based method.

    Parameters:
        dynmat_crystal (np.array): Dynamical matrix in crystal space.
        grid (np.array): Grid points.
        nrep (int): Replication factor.
        grid_type (str): 'F' for Fortran-like, 'C' for C-like ordering.

    Returns:
        np.array: Dynamical matrix in real space.
    """
    tmp = grid[np.newaxis, :] - grid[:, np.newaxis]
    mask = find_index_from_grid(grid=tmp, nrep=nrep, grid_type=grid_type)
    dynmat_R_R = dynmat_crystal[0, :, :, mask, :, :]
    dynmat_R_R = np.transpose(dynmat_R_R, axes=[1, 2, 3, 0, 4, 5])
    return dynmat_R_R


def sparse_from_crystal2real_space_fast(
    dynmat_crystal, grid, nrep, grid_type="F", threshold=1e-14
):
    """
    Convert the crystal-space dynamical matrix to a sparse real-space representation.

    Parameters:
        dynmat_crystal (np.array): Real part of the crystal-space dynamical matrix.
        grid (np.array): Grid array.
        nrep (int): Replication factor.
        grid_type (str): Grid ordering ('F' or 'C').
        threshold (float): Threshold below which values are set to zero.

    Returns:
        scipy.sparse.csr_matrix: Sparse representation of the dynamical matrix.
    """
    dynmat_R_R_sparse = []
    n_k = len(grid)
    n_modes = dynmat_crystal.shape[1] * 3
    for i in range(n_k):
        tmp = grid[i] - grid[:]
        mask = find_index_from_grid(grid=tmp, nrep=nrep, grid_type=grid_type)
        dynmat_R_ = dynmat_crystal[0, :, :, mask, :, :]
        dynmat_R_ = np.transpose(dynmat_R_, axes=(1, 2, 0, 3, 4))
        dynmat_R_ = dynmat_R_.reshape([1, n_k * n_modes * n_modes])
        dynmat_R_ = np.where(np.abs(dynmat_R_) > threshold, dynmat_R_, 0)
        dynmat_R_R_sparse.append(sp.csr_matrix(dynmat_R_))
    dynmat_R_R_sparse = sp.vstack(dynmat_R_R_sparse)
    return dynmat_R_R_sparse


def from_crystal2real_space(
    dynmat_crystal, replicated_cell, list_of_replicas, tol=1e-6
):
    """
    Convert crystal-space dynamical matrix to a full real-space version.

    Parameters:
        dynmat_crystal (np.array): Crystal-space dynamical matrix.
        replicated_cell (np.array): Matrix of the replicated cell.
        list_of_replicas (np.array): Array of replica positions.
        tol (float): Tolerance for position matching.

    Returns:
        np.array: Full real-space dynamical matrix.
    """
    cell_inv = np.linalg.inv(replicated_cell)
    nq = dynmat_crystal.shape[3]
    n_basis = dynmat_crystal.shape[1]
    logging.warning("n replicas = " + str(nq))
    dynmat = np.zeros([nq, n_basis, 3, nq, n_basis, 3], dtype=complex)
    for iq in range(nq):
        for jq in range(nq):
            a = wrap_coordinates(
                list_of_replicas[iq] - list_of_replicas[jq],
                cell=replicated_cell,
                cell_inv=cell_inv,
            )
            index = np.where(np.linalg.norm(list_of_replicas - a, axis=1) < tol)
            if np.prod(index[0].shape) == 1:
                dynmat[iq, :, :, jq, :, :] = dynmat_crystal[0, :, :, index, :, :]
            else:
                print("Index not uniquely found for:", a, index)
    return dynmat


def wrap_coordinates(dxij, cell=None, cell_inv=None, tol=1e-6):
    """
    Wrap coordinates to calculate the shortest distance exploiting periodicity.

    Parameters:
        dxij (np.array): Coordinate differences.
        cell (np.array, optional): Cell matrix for transformation.
        cell_inv (np.array, optional): Inverse cell matrix.
        tol (float): Tolerance to avoid ambiguity.

    Returns:
        np.array: Wrapped coordinate differences.
    """
    if cell is not None and cell_inv is None:
        cell_inv = np.linalg.inv(cell)
    if cell is not None:
        dxij = dxij.dot(cell_inv)
    dxij = dxij - np.round(dxij - tol)
    if cell is not None:
        dxij = dxij.dot(cell)
    return dxij


def second2dynmat(second, mass):
    """
    Convert second-order force constants to dynamical matrix.

    Parameters:
        second (np.array): Second-order force constants.
        mass (np.array): Array of masses.

    Returns:
        np.array: Dynamical matrix.
    """
    dynmat = second * 1 / np.sqrt(mass[:, np.newaxis, np.newaxis, np.newaxis])
    dynmat = dynmat * 1 / np.sqrt(mass[np.newaxis, np.newaxis, :, np.newaxis])
    evtotenjovermol = units.mol / (10 * units.J)
    dynmat *= evtotenjovermol
    return dynmat


def dynmat2second(dynmat, mass):
    """
    Convert dynamical matrix back to second-order force constants.

    Parameters:
        dynmat (np.array): Dynamical matrix.
        mass (np.array): Array of masses.

    Returns:
        np.array: Second-order force constants.
    """
    second = dynmat * np.sqrt(mass[:, np.newaxis, np.newaxis, np.newaxis])
    second = second * np.sqrt(mass[np.newaxis, np.newaxis, :, np.newaxis])
    evtotenjovermol = units.mol / (10 * units.J)
    second /= evtotenjovermol
    return second


def grid_repeat(unit_atoms, supercell, grid_type="F"):
    """
    Repeat the unit cell to form a supercell.

    Parameters:
        unit_atoms (ase.atoms.Atoms): Single cell atoms.
        supercell (list): Multiplicity factors along each axis.
        grid_type (str): 'F' for Fortran-like ordering or 'C' for C-like ordering.

    Returns:
        ase.atoms.Atoms: Supercell with repeated atoms.
    """
    if grid_type == "F":
        replicated_atoms = (
            unit_atoms.copy()
            * (supercell[0], 1, 1)
            * (1, supercell[1], 1)
            * (1, 1, supercell[2])
        )
    elif grid_type == "C":
        replicated_atoms = (
            unit_atoms.copy()
            * (1, 1, supercell[2])
            * (1, supercell[1], 1)
            * (supercell[0], 1, 1)
        )
    else:
        print("Grid type not acceptable")
        return
    return replicated_atoms


def grid_F2C(grid):
    """
    Convert grid from Fortran to C ordering.

    Parameters:
        grid (np.array): Input grid.

    Returns:
        np.array: Grid with transposed axes.
    """
    if len(grid.shape) == 3:
        grid = np.transpose(grid, axes=(2, 1, 0))
    elif len(grid.shape) == 4:
        grid = np.transpose(grid, axes=(0, 3, 2, 1))
    return grid


# Main computation
for c in [0.5]:
    ini_sc = [7, 7, 7]
    root = "../../0_DFT/c_{}/fc_kaldo_sc{}/".format(c, ini_sc[0])
    forceconstants = ForceConstants.from_folder(
        root, supercell=ini_sc, format="shengbte-d3q", only_second=True
    )

    n_replicas, n_unit_atoms, _ = forceconstants.second.replicated_positions.shape
    detected_grid = np.round(
        (
            forceconstants.second.replicated_positions.reshape(
                (n_replicas, n_unit_atoms, 3)
            )
            - forceconstants.second.atoms.positions[np.newaxis, :, :]
        ).dot(np.linalg.inv(forceconstants.second.atoms.cell))[:, 0, :],
        0,
    ).astype(np.int)

    grid_c = Grid(grid_shape=forceconstants.second.supercell, order="C")
    grid_fortran = Grid(grid_shape=forceconstants.second.supercell, order="F")
    if (grid_c.grid(is_wrapping=False) == detected_grid).all():
        grid_type = "C"
        logging.debug("Using C-style position grid")
    elif (grid_fortran.grid(is_wrapping=False) == detected_grid).all():
        grid_type = "F"
        print("Using Fortran-style position grid")
    else:
        err_msg = "Unable to detect grid type"
        logging.error(err_msg)
        raise TypeError(err_msg)

    n_modes = 6
    temperature = 300
    k_points = int(sys.argv[1])
    phonons_config = {
        "kpts": [k_points, k_points, k_points],
        "is_classic": False,
        "temperature": temperature,
        "folder": "../../1_crystal/ald_c_{}_sc{}".format(c, ini_sc[0]),
        "storage": "numpy",
    }
    # Setup phonon object with configuration details and the computed force constants
    phonons = Phonons(forceconstants=forceconstants, **phonons_config)
    freq = np.sort(phonons.frequency.flatten())

    q_points = phonons._reciprocal_grid.unitary_grid(is_wrapping=False)
    frequency = np.zeros((phonons.n_k_points, phonons.n_modes))
    dynmat_q_list = []
    grid = Grid([k_points, k_points, k_points], order=grid_type).grid(is_wrapping=True)
    list_of_replicas = grid.dot(phonons.atoms.cell)
    for ik in range(len(q_points)):
        q_point = q_points[ik]
        phonon = HarmonicWithQ(
            q_point=q_point,
            second=phonons.forceconstants.second,
            distance_threshold=phonons.forceconstants.distance_threshold,
            folder=phonons.folder,
            storage=phonons.storage,
            is_nw=phonons.is_nw,
            is_unfolding=phonons.is_unfolding,
        )
        dynmat_q_list.append(phonon._dynmat_fourier.numpy())
    dynmat_q_list = np.array(dynmat_q_list)

    cell_inv = np.linalg.inv(phonons.atoms.cell)
    start = time()
    dynmat_q_fft = dynmat_q_list.reshape(
        [k_points, k_points, k_points, n_modes, n_modes]
    )
    dynmat_R_fft = np.zeros_like(dynmat_q_fft)
    n_k = len(q_points)
    for s1 in range(n_modes):
        for s2 in range(n_modes):
            dynmat_R_fft[:, :, :, s1, s2] = (
                FFT(grid_F2C(dynmat_q_fft[:, :, :, s1, s2])) / n_k
            )
    dynmat_R_fft = dynmat_R_fft.reshape([n_k, n_modes, n_modes])
    dynmat_R_fft = dynmat_R_fft.reshape([1, n_k, 2, 3, 2, 3])
    dynmat_R = np.transpose(dynmat_R_fft, axes=(0, 2, 3, 1, 4, 5))
    end = time()
    logging.info("fft q to r time={}".format(end - start))

    # Create sparse second-order force constants
    n_k = len(q_points)
    replicated_atoms = grid_repeat(
        unit_atoms=forceconstants.atoms,
        supercell=[k_points, k_points, k_points],
        grid_type=grid_type,
    )
    n_atoms = replicated_atoms.positions.shape[0]
    n_def = int(n_atoms * c)
    mass = replicated_atoms.get_masses()
    evtotenjovermol = units.mol / (10 * units.J)
    start = time()

    second_sparse = sparse_from_crystal2real_space_fast(
        dynmat_crystal=np.real(dynmat_R) * mass[0] / evtotenjovermol,
        grid=grid,
        nrep=k_points,
        grid_type=grid_type,
        threshold=1e-14,
    )
    second_sparse = second_sparse.reshape([n_k * n_modes, n_k * n_modes])
    end = time()
    logging.info("translational invariance time={}".format(end - start))

    sparseness = 1 - second_sparse.count_nonzero() / np.prod(second_sparse.shape)
    logging.info("sparseness={}".format(sparseness))
    Si = Atoms("Si")
    Si_mass = 28.086
    Ge = Atoms("Ge")
    Ge_mass = Ge.get_masses()[0]

    new_mass = np.ones_like(mass) * Si_mass

    np.random.seed(seed=687578)
    list_def = random.sample(range(0, n_atoms), n_def)
    list_def = np.sort(np.array(list_def))

    print("Si and Ge mass= ", Si_mass, " ", Ge_mass)
    formula = replicated_atoms.get_chemical_symbols()
    for i in list_def:
        formula[i] = "Ge"
        new_mass[i] = Ge_mass

    start = time()
    new_root = "./c_{}/N{}/0/fc/".format(c, k_points)
    if not os.path.exists(new_root):
        os.makedirs(new_root)
    sp.save_npz(new_root + "second.npz", second_sparse)
    end = time()
    logging.info("saving time={}".format(end - start))
    final_atoms = replicated_atoms.copy()
    final_atoms.set_chemical_symbols(formula)
    final_atoms.set_masses(new_mass)
    write(new_root + "replicated_atoms.xyz", final_atoms)

    if k_points < 19:
        second = second_sparse.toarray()
        np.save(new_root + "second.npy", second.reshape([1, n_atoms, 3, 1, n_atoms, 3]))

    # Other mass configuration samples
    nsample = 4
    print("generating other mass configurations")
    for isample in range(1, nsample):
        new_mass = np.ones_like(mass) * Si_mass
        np.random.seed(seed=687578)
        list_def = random.sample(range(0, n_atoms), n_def)
        list_def = np.sort(np.array(list_def))

        print("Si and Ge mass= ", Si_mass, " ", Ge_mass)
        formula = replicated_atoms.get_chemical_symbols()
        for i in list_def:
            formula[i] = "Ge"
            new_mass[i] = Ge_mass

        new_root = "./c_{}/N{}/{}/fc/".format(c, k_points, isample)
        if not os.path.exists(new_root):
            os.makedirs(new_root)
        final_atoms = replicated_atoms.copy()
        final_atoms.set_chemical_symbols(formula)
        final_atoms.set_masses(new_mass)
        write(new_root + "replicated_atoms.xyz", final_atoms)
