import numpy as np
from scipy.fft import fftn, ifftn
from ase.io import read, write
from multiprocessing import Pool
import os
import sys
from time import time

# Get number of CPUs available
ncpus = os.cpu_count()
print(ncpus)

# Discrete Monte Carlo minimization
# 3D case

dimension = 3
if dimension == 1:
    from scipy.fft import fft as FFT, ifft as IFFT
elif dimension == 3:
    # Use n-dimensional FFT in 3D
    FFT = fftn
    IFFT = ifftn


def loss(Cq, phiq, norm_loss=None):
    """
    Compute loss as the norm of the difference between |Cq| and phiq * conj(phiq).
    """
    diff = np.abs(Cq) - phiq * np.conjugate(phiq)
    if norm_loss is not None:
        return np.linalg.norm(diff) / norm_loss
    return np.linalg.norm(diff)


def update(vector):
    """
    Randomly swap two elements of the flattened vector,
    recalculate its FFT, and verify the inverse transform.
    """
    original_shape = vector.shape
    vector = vector.flatten()
    L = len(vector)
    i, j = np.random.randint(0, L, size=2)
    tmp_vector = vector.copy()
    tmp_vector[i] = vector[j]
    tmp_vector[j] = vector[i]
    tmp_vector = tmp_vector.reshape(original_shape)
    new_phiq = FFT(tmp_vector)
    error = np.linalg.norm(tmp_vector - IFFT(new_phiq))
    if error > 1e-6:
        print("New vector and its FT do not coincide!")
    return tmp_vector, new_phiq


def mc_move(mq0, vector0, Cq, beta=5, norm_loss=None):
    """
    Perform a Monte Carlo update move using the Metropolis criterion.
    """
    l0 = loss(Cq, phiq=mq0, norm_loss=norm_loss)
    vector1, mq1 = update(vector=vector0)
    l1 = loss(Cq, phiq=mq1, norm_loss=norm_loss)
    r = np.random.rand(1)
    logratio = beta * (np.log(l0) - np.log(l1))
    ratio = np.exp(logratio)
    if r < ratio:
        return vector1, mq1, l1
    return vector0, mq0, l0


def rphase_symmetrization3D(q_list, rphase):
    """
    Enforce conjugate symmetry for random phase array.
    """
    tmp = rphase.copy()
    count = 0
    for iq, q_ in enumerate(q_list):
        mask = np.where(np.linalg.norm(q_list + q_, axis=1) < 1e-8)
        if len(mask[0]) == 1:
            tmp[mask] = -tmp[iq]
        if len(mask[0]) == 0:
            tmp[iq] = 0
            count += 1
    print("q vectors without conjugate:", count)
    mask = np.where(np.linalg.norm(q_list, axis=1) < 1e-8)
    tmp[mask] = 0
    return tmp


def reshape_to_grid(Cr, grid, dim=3, mode="direct"):
    """
    Reshape Cr to match the provided grid.
    """
    if mode == "direct":
        Cr_ = Cr[:, np.newaxis] * np.ones(dim)
        Cr_ = Cr_.T
        Cr_ = Cr_.reshape(grid.shape)
    elif mode == "inverse":
        L = grid.shape[-1]
        Cr_ = Cr.reshape([dim, L**dim]).T
    return Cr_


def compute_Cq(q_list, sigma=1, method="gauss", reg=1e-4):
    """
    Compute the Fourier space configuration amplitude Cq.
    """
    dim_index = np.where((np.array(q_list.shape) - 3) == 0)[0][0]
    q_norm = np.linalg.norm(q_list, axis=dim_index)
    if method == "gauss":
        Cq = np.exp(-((q_norm * sigma) ** 2) / 2)
        Cq *= np.sqrt(2 * np.pi) * sigma**3
    elif method == "corr":
        Cq = 1 / (np.linalg.norm(q_list + 1j * reg, axis=dim_index)) ** 2
    return Cq


def gauss(w, eta):
    """
    Return a Gaussian function evaluated at w with width eta.
    """
    return 1 / np.sqrt(2 * np.pi * eta**2) * np.exp(-(w**2) / (2 * eta**2))


def gauss_kernel(kappa, w, new_w, eta):
    """
    Compute the Gaussian kernel convolution.
    """
    G = gauss(new_w[:, np.newaxis] - w[np.newaxis, :], eta)
    num = np.einsum("n,in->i", kappa, G)
    return num


def chi(qvec, list_of_replicas, cell_inv):
    """
    Compute the chi function for a given q vector and list of replicas.
    """
    chi_k = np.exp(1j * 2 * np.pi * list_of_replicas.dot(cell_inv.dot(qvec.T)))
    return chi_k


def grid_repeat(unit_atoms, supercell, grid_type="F"):
    """
    Repeat unit cell atoms into a supercell according to grid_type.
    """
    if grid_type == "F":
        replicated_atoms = (
            unit_atoms.copy()
            * (supercell[0], 1, 1)
            * (1, supercell[1], 1)
            * (1, 1, supercell[2])
        )
    elif grid_type == "C":
        replicated_atoms = (
            unit_atoms.copy()
            * (1, 1, supercell[2])
            * (1, supercell[1], 1)
            * (supercell[0], 1, 1)
        )
    else:
        print("Grid type not acceptable")
        return
    return replicated_atoms


def find_index_from_grid(grid, nrep, grid_type="F"):
    """
    Find unique index from grid coordinates.
    """
    grid = np.mod(grid, nrep)
    if grid_type == "F":
        mask = grid * np.array([1, nrep, nrep**2])
        mask = np.sum(mask, axis=-1)
    if grid_type == "C":
        mask = grid * np.array([nrep**2, nrep, 1])
        mask = np.sum(mask, axis=-1)
    return mask


def grid_F2C(grid):
    """
    Convert grid from F-type to C-type ordering.
    """
    if len(grid.shape) == 3:
        grid = np.transpose(grid, axes=(2, 1, 0))
    elif len(grid.shape) == 4:
        grid = np.transpose(grid, axes=(0, 3, 2, 1))
    return grid


def compute_gr_cell(i, types, positions, cell, r, delta):
    """
    Compute the Gaussian-distributed radial function for cell i.
    """
    pos_i = positions[i]
    dist_i = positions - pos_i
    dist_i = wrap_coordinates(dist_i, cell=cell)
    dist_i = np.linalg.norm(dist_i, axis=1)
    mask = np.ones_like(types)
    g_r = gauss_kernel(mask, w=dist_i, new_w=r, eta=delta)
    return g_r


def compute_corr_cell(i, types, positions, cell, r, delta):
    """
    Compute the correlation function for cell i.
    """
    pos_i = positions[i]
    dist_i = positions - pos_i
    dist_i = wrap_coordinates(dist_i, cell=cell)
    dist_i = np.linalg.norm(dist_i, axis=1)
    corr = types[i] * types
    corr_r = gauss_kernel(corr, w=dist_i, new_w=r, eta=delta)
    return corr_r


def wrap_coordinates(dxij, cell=None, cell_inv=None, tol=1e-6):
    """
    Wrap coordinates using periodic boundary conditions to ensure minimal distance.
    """
    if cell is not None and cell_inv is None:
        cell_inv = np.linalg.inv(cell)
    if cell is not None:
        dxij = dxij.dot(cell_inv)
    dxij = dxij - np.round(dxij - tol)
    if cell is not None:
        dxij = dxij.dot(cell)
    return dxij


class correlated_disorder:
    """
    Class to generate correlated disorder configurations.
    """

    def __init__(
        self,
        L,
        dim=3,
        method="gauss",
        sigma=1,
        use_mc=True,
        mc_steps=2e5,
        set_zero=False,
        beta=10,
    ):
        self.L = L
        self.dim = dim
        self.method = method
        self.sigma = sigma
        self.use_mc = use_mc
        self.mc_steps = mc_steps
        self.set_zero = set_zero
        if not method == "gauss":
            self.set_zero = True
        if dim == 3:
            grid = np.mgrid[0:L, 0:L, 0:L]
        qmesh = (grid - L * np.round(grid / L)) * 2 * np.pi / L
        q_list = qmesh.reshape(dim, L**dim).T
        self.qmesh = qmesh
        self.q_list = q_list
        pos_flat = (grid - L * np.round(grid / L)).reshape([dim, L**dim]).T
        self.R_adi = np.linalg.norm(pos_flat, axis=1)  # dimensionless distances
        self.Cq = None
        self.rphase = np.random.rand(L**dim) * 2 * np.pi
        self.mq0 = None
        self.beta = beta
        self.norm_loss = None

    def generate_configuration(self):
        """
        Generate a configuration by computing Cq and applying Monte Carlo moves.
        """
        Cq = compute_Cq(q_list=self.qmesh, sigma=self.sigma, method=self.method)
        print(Cq.shape)
        self.Cq = Cq
        Cr = IFFT(Cq)
        if self.set_zero:
            Cr = Cr - np.min(Cr)
            Cq = FFT(Cr)
            self.Cq = Cq
        Vq = np.sqrt(np.abs(Cq))
        rphase_sym = rphase_symmetrization3D(self.q_list, self.rphase)
        rphase_sym = rphase_sym.reshape(Vq.shape)
        Vq = Vq * np.exp(1j * rphase_sym)
        # Generate initial vector
        Vr = IFFT(Vq)
        print("Vr mean and std =", np.mean(Vr), np.std(Vr))
        disc = np.sign(Vr - np.mean(Vr))
        disc = disc * (np.std(Vr) / np.std(disc))
        vector = disc
        mq = FFT(vector)
        self.mq0 = mq
        if not self.use_mc:
            return mq
        else:
            loss_list = []
            norm_loss = loss(Cq, phiq=self.mq0)
            self.norm_loss = norm_loss
            print("Initial loss:", norm_loss)
            beta = self.beta
            mc_steps = self.mc_steps
            for i in range(int(mc_steps)):
                if np.mod(i, 200) == 0:
                    beta *= 1.01
                vector, mq, ell = mc_move(
                    mq0=mq, vector0=vector, Cq=Cq, beta=beta, norm_loss=norm_loss
                )
                loss_list.append(ell)
            return mq, loss_list


# Main script execution
L = int(sys.argv[1])
mc_steps = 4e5
if L < 20:
    mc_steps = 2e5

seed_list = [687578, 736786, 415189, 903892]  # random seeds extracted from time digits
for isample, seed in enumerate(seed_list):
    np.random.seed(seed=seed)
    system = correlated_disorder(L=L, method="corr", sigma=1.25, mc_steps=mc_steps)
    mq, loss_list = system.generate_configuration()

    bias_loss = np.abs(system.Cq[0, 0, 0]) / system.norm_loss

    # Load atomic configuration
    atoms = read("../../1_crystal/unit_configs/unit_atoms_c_0.5.xyz")
    cell = atoms.cell
    supercell = [L, L, L]
    cell_inv = np.linalg.inv(cell)
    replicated_atoms = grid_repeat(atoms, supercell, grid_type="F")
    pos_bravais = replicated_atoms.positions.dot(cell_inv).reshape([L**3, 2, 3])

    mr = IFFT(mq)
    err = np.linalg.norm(np.imag(mr))
    if err < 1e-8:
        print("mr is real as it should be")
        mr = np.real(mr)
    else:
        print("Error =", err)
    mr = grid_F2C(mr)
    unit_basis = 2
    types = np.sign(mr).flatten()[:, np.newaxis] * np.ones(unit_basis)[np.newaxis, :]
    types = types.flatten()

    # Plot spatial correlation (calculation only, no plot shown)
    delta = 0.5
    positions = replicated_atoms.positions
    L_avg = (replicated_atoms.get_volume()) ** (1 / 3)
    r = np.linspace(0.5, L_avg / 2, 200)
    n_atoms = len(types)
    g_r = np.zeros_like(r)
    corr_r = np.zeros_like(r)
    rep_cell = replicated_atoms.cell
    start = time()

    with Pool(ncpus) as p:
        inputs = [(i, types, positions, rep_cell, r, delta) for i in range(n_atoms)]
        result = p.starmap(compute_gr_cell, inputs)
    g_r = np.mean(np.array(result), axis=0)

    with Pool(ncpus) as p:
        inputs = [(i, types, positions, rep_cell, r, delta) for i in range(n_atoms)]
        result = p.starmap(compute_corr_cell, inputs)
    end = time()
    print("Time to compute corr and gdr:", end - start)
    corr_r = np.mean(np.array(result), axis=0)

    # Save configuration
    new_atoms = replicated_atoms.copy()
    Si_mass = 28.086  # ASE gives 28.085
    Ge_mass = 72.64
    new_mass = np.where(types == 1, Ge_mass, Si_mass)
    new_formula = np.where(types == 1, "Ge", "Si").tolist()
    new_atoms.set_chemical_symbols(new_formula)
    new_atoms.set_masses(new_mass)
    new_root = "c_0.5/"
    if system.method == "corr":
        new_root = new_root + "corr_N{}/{}/fc/".format(system.L, isample)
    elif system.method == "gauss":
        new_root = new_root + "gauss_sigma{}_N{}/{}/fc/".format(
            system.sigma, system.L, isample
        )
    if not os.path.exists(new_root):
        os.makedirs(new_root)
    write(new_root + "replicated_atoms.xyz", new_atoms)
    system_info = {
        "loss": loss_list,
        "bias_loss": bias_loss,
        "gdr": g_r,
        "corr_r": corr_r,
        "r": r,
    }
    np.save(new_root + "system_info.npy", system_info)
