import numpy as np
import sys
import logging

from kaldo.forceconstants import ForceConstants
from kaldo.phonons import Phonons
from kaldo.grid import Grid
import ase.units as units


def dynmat2second(dynmat, mass):
    """
    Convert a dynamical matrix to second order force constants.

    Parameters:
      dynmat : array
          Dynamical matrix.
      mass : array
          Atomic masses.

    Returns:
      second : array
          Second order force constants.
    """
    second = dynmat * np.sqrt(mass[:, np.newaxis, np.newaxis, np.newaxis])
    second = second * np.sqrt(mass[np.newaxis, np.newaxis, :, np.newaxis])
    evtotenjovermol = units.mol / (10 * units.J)
    second /= evtotenjovermol
    return second


def second2dynmat(second, mass):
    """
    Convert second order force constants back to a dynamical matrix.

    Parameters:
      second : array
          Second order force constants.
      mass : array
          Atomic masses.

    Returns:
      dynmat : array
          Dynamical matrix.
    """
    dynmat = second / np.sqrt(mass[:, np.newaxis, np.newaxis, np.newaxis])
    dynmat = dynmat / np.sqrt(mass[np.newaxis, np.newaxis, :, np.newaxis])
    evtotenjovermol = units.mol / (10 * units.J)
    dynmat *= evtotenjovermol
    return dynmat


c = float(sys.argv[1])
if c == 0 or c == 1:
    c = int(c)
k_points = int(sys.argv[2])
ini_sc = [7, 7, 7]
root = "../../../../../0_DFT/" + "c_{}/fc_kaldo_sc{}/".format(c, ini_sc[0])
print("eigvectors taken from: ", root)
forceconstants = ForceConstants.from_folder(
    root, supercell=ini_sc, format="shengbte-d3q", only_second=True
)

temperature = 300
phonons_config = {
    "kpts": [k_points, k_points, k_points],
    "is_classic": False,
    "temperature": temperature,  # temperature = 300K
    "folder": "../../../../../1_crystal/ald_c_{}_sc{}".format(c, ini_sc[0]),
    "storage": "numpy",
}
# Set up phonon object with configuration and force constants
phonons = Phonons(forceconstants=forceconstants, **phonons_config)
q_points = phonons._reciprocal_grid.unitary_grid(is_wrapping=False)
eig = phonons.eigenvectors
w2 = phonons.eigenvalues

# Determine grid type
n_replicas, n_unit_atoms, _ = forceconstants.second.replicated_positions.shape
detected_grid = np.round(
    (
        forceconstants.second.replicated_positions.reshape(
            (n_replicas, n_unit_atoms, 3)
        )
        - forceconstants.second.atoms.positions[np.newaxis, :, :]
    ).dot(np.linalg.inv(forceconstants.second.atoms.cell))[:, 0, :],
    0,
).astype(np.int)
grid_c = Grid(grid_shape=forceconstants.second.supercell, order="C")
grid_fortran = Grid(grid_shape=forceconstants.second.supercell, order="F")
if (grid_c.grid(is_wrapping=False) == detected_grid).all():
    grid_type = "C"
    logging.debug("Using C-style position grid")
elif (grid_fortran.grid(is_wrapping=False) == detected_grid).all():
    grid_type = "F"
    print("Using fortran-style position grid")
else:
    err_msg = "Unable to detect grid type"
    logging.error(err_msg)
    raise TypeError(err_msg)

mesh = int(k_points / 2)
k_path = [
    np.array([mesh / k_points, mesh / k_points, mesh / k_points])
    - np.array([1, 1, 1]) * (i / k_points)
    for i in range(1, mesh)
]
k_path += [np.array([0, 1, 1]) * i / k_points for i in range(mesh)]
grid = Grid([k_points, k_points, k_points], order=grid_type).grid(is_wrapping=True)
list_of_replicas = grid.dot(phonons.atoms.cell)
n_modes = phonons.n_modes
n_phonons = phonons.n_phonons
v_array = np.zeros([len(k_path), n_modes, n_phonons], dtype=complex)
f_array = np.zeros([len(k_path), n_modes])
my_path = {"q": k_path}

for ik, q_point in enumerate(k_path):
    ik_map = np.where(np.linalg.norm(q_point - q_points, axis=1) < 1e-6)[0]
    for imod in range(n_modes):
        v_ = eig[ik_map, :, imod].reshape(int(n_modes / 3), 3)
        v_ext = np.einsum(
            "R,ix->Rix", np.exp(-1j * 2 * np.pi * np.dot(grid, q_point)), v_
        ).reshape([len(q_points) * 2, 3])
        v_array[ik, imod, :] = v_ext.flatten() / np.linalg.norm(v_ext)

my_path["eig"] = v_array
np.save("eig_path.npy", my_path)
