import numpy as np
import sys
import logging
import os
from time import time
from scipy import sparse
from multiprocessing import Pool

from ase.io import read
from ase import units

from hydro_glasses import lanczos

logging.basicConfig(level=logging.INFO)


def second2dynmat_sparse(second_sparse, mass):
    """
    Compute the dynamical matrix as a sparse matrix from the second derivative matrix
    and atomic masses.

    Parameters:
        second_sparse (sparse matrix): Second derivative matrix.
        mass (ndarray): Array of atomic masses.

    Returns:
        sparse matrix: Dynamical matrix.
    """
    # Conversion factor from energy units
    evtotenjovermol = units.mol / (10 * units.J)
    sqrtm = 1 / np.sqrt((mass[:, np.newaxis] * np.ones(3)[np.newaxis, :]).flatten())
    sqrtm = sparse.diags(sqrtm)
    sqrtm_sparse = sparse.csr_matrix(sqrtm)
    dynmat_sparse = sqrtm_sparse @ (second_sparse @ sqrtm_sparse)
    dynmat_sparse = dynmat_sparse * evtotenjovermol
    return dynmat_sparse


def lorentzian(w, w0, tau, norm):
    """
    Evaluate a Lorentzian function.

    Parameters:
        w (float or ndarray): Frequency variable.
        w0 (float): Center frequency.
        tau (float): Broadening parameter.
        norm (float): Normalization factor.

    Returns:
        float or ndarray: Lorentzian value at w.
    """
    # Alternative version was commented out in original code.
    return norm * (tau / (tau**2 + (w - w0) ** 2))


def DHO(w, w0, tau, norm):
    """
    Evaluate the response of a damped harmonic oscillator (DHO).

    Parameters:
        w (float or ndarray): Frequency variable.
        w0 (float): Resonance frequency.
        tau (float): Damping parameter.
        norm (float): Normalization factor.

    Returns:
        float or ndarray: DHO response at w.
    """
    return norm * (w * 2 * tau / ((w * tau) ** 2 + (w**2 - w0**2) ** 2))


def lor_2(w, gamma, norm, w0, eta):
    """
    Evaluate a modified Lorentzian function.

    Parameters:
        w (float or ndarray): Frequency variable.
        gamma (float): Damping parameter.
        norm (float): Normalization factor.
        w0 (float): Center frequency.
        eta (float): Additional broadening parameter.

    Returns:
        float or ndarray: Modified Lorentzian value at w.
    """
    w0_complex = w0 + 1j * np.abs(gamma)
    y = w0_complex * eta / ((w**2 - w0_complex**2) ** 2 + (eta**2) * (w0_complex**2))
    return norm * np.imag(1j * y)


def DHO_2(w, gamma, norm, w0, eta):
    """
    Evaluate an alternative damped harmonic oscillator (DHO) function.

    Parameters:
        w (float or ndarray): Frequency variable.
        gamma (float): Damping parameter.
        norm (float): Normalization factor.
        w0 (float): Resonance frequency.
        eta (float): Additional broadening parameter.

    Returns:
        float or ndarray: Alternative DHO response at w.
    """
    gamma = np.abs(gamma)
    num = norm * (2 * gamma + eta) * w0
    den = (w**2 - w0**2 - 0.5 * (gamma + eta) ** 2) ** 2 + w**2 * (2 * gamma + eta) ** 2
    return num / den


# Input parameter for Lanczos method from command-line arguments
eta = float(sys.argv[1])
ortho = False
k = 100  # Lanczos chain length
root = "fc/"
logging.info(root)

# Read atomic positions and masses
atoms = read(root + "replicated_atoms.xyz")
mass = atoms.get_masses()

threshold = 1e-14
try:
    # Attempt to load the sparse dynamical matrix if it exists
    dynmat_sparse = sparse.load_npz(root + "dynmat.npz")
    # Ensure symmetry for numerical reasons
    dynmat_sparse = (dynmat_sparse + dynmat_sparse.transpose()) / 2
except Exception as e:
    logging.info("Exception: " + str(e))
    try:
        # Load second derivative matrix and compute dynamical matrix
        second_sparse = sparse.load_npz(root + "second.npz")
        dynmat_sparse = second2dynmat_sparse(second_sparse=second_sparse, mass=mass)
        # Ensure symmetry for numerical reasons
        dynmat_sparse = (dynmat_sparse + dynmat_sparse.transpose()) / 2
        sparse.save_npz(root + "dynmat.npz", dynmat_sparse)
    except Exception as e:
        logging.info("Exception: " + str(e))
        # Fallback for non-sparse matrix
        dynmat = np.load(root + "second.npy")
        dynmat = np.where(np.abs(dynmat) > threshold, dynmat, 0)
        sparsity = 1.0 - np.count_nonzero(dynmat) / dynmat.size
        logging.info("sparsity " + str(sparsity))

        # Normalize by masses
        dynmat = dynmat / np.sqrt(
            mass[np.newaxis, :, np.newaxis, np.newaxis, np.newaxis, np.newaxis]
        )
        dynmat = dynmat / np.sqrt(
            mass[np.newaxis, np.newaxis, np.newaxis, np.newaxis, :, np.newaxis]
        )
        n_modes = dynmat.shape[1] * dynmat.shape[2]
        logging.info("nmodes= " + str(n_modes))
        dynmat *= units.mol / (10 * units.J)
        dynmat = dynmat.reshape([n_modes, n_modes])

        # Ensure symmetry
        dynmat = (dynmat + dynmat.T) / 2
        dynmat_sparse = sparse.csr_matrix(dynmat)
        sparse.save_npz(root + "dynmat.npz", dynmat_sparse)


# Load eigenvectors for the virtual crystal
my_path = np.load("eig_path.npy", allow_pickle=True).item()
n_modes = 6
Q_array = np.array(my_path["q"])
ket = np.transpose(my_path["eig"], axes=[2, 0, 1])

# Define frequency array for spectrum calculation
spectrum = {}
omega_array = np.linspace(0.1, 120, 10000)
spectrum["omega"] = omega_array
spectrum["Q"] = Q_array

ncpus = os.cpu_count()
logging.info("ncpus:{}".format(ncpus))
if ncpus > 20:
    ncpus = 8

result = []
for i_mod in range(n_modes):
    start = time()
    with Pool(ncpus) as p:
        inputs = [
            (dynmat_sparse, ket[:, i, i_mod], k, omega_array, eta, True, ortho)
            for i in range(ket.shape[1])
        ]
        result_ = p.starmap(lanczos.spectrum, inputs)
    end = time()
    logging.info("First polarization done in {} s".format(end - start))
    result.append(result_)
result = np.array(result)

print(result.shape)

# Organize results into spectrum dictionary
for iq in range(Q_array.shape[0]):
    spectrum[iq] = {}
    for ib in range(n_modes):
        spectrum[iq][ib] = {}
        spectrum[iq][ib]["S"] = result[ib, iq, 0]
        spectrum[iq][ib]["alpha"] = result[ib, iq, 1]
        spectrum[iq][ib]["beta"] = result[ib, iq, 2]

root_spec = "./spectrum/"
if not os.path.exists(root_spec):
    os.makedirs(root_spec)
if ortho:
    np.save(root_spec + "spectrum_path_ortho_k{}_eta{}.npy".format(k, eta), spectrum)
else:
    np.save(root_spec + "spectrum_path_k{}_eta{}.npy".format(k, eta), spectrum)
