import numpy as np
import sys
import os
import logging
from time import time
from ase.io import read
from ase import units
from scipy import sparse
from hydro_glasses import lanczos

logging.basicConfig(level=logging.INFO)


def second2dynmat_sparse(second_sparse, mass):
    """
    Convert the second derivative sparse matrix to a dynamical matrix.

    Parameters:
        second_sparse (scipy.sparse matrix): Second derivative matrix (square matrix).
        mass (numpy.ndarray): Array of atomic masses.

    Returns:
        scipy.sparse matrix: Symmetrized dynamical matrix.
    """
    evtotenjovermol = units.mol / (10 * units.J)
    sqrtm = 1 / np.sqrt((mass[:, np.newaxis] * np.ones(3)[np.newaxis, :]).flatten())
    sqrtm = np.diag(sqrtm)
    sqrtm_sparse = sparse.csr_matrix(sqrtm)
    dynmat_sparse = sqrtm_sparse @ (second_sparse @ sqrtm_sparse)
    dynmat_sparse = dynmat_sparse * evtotenjovermol
    return dynmat_sparse


def levi_civita(x, y, z):
    """
    Compute the Levi-Civita symbol for three indices.

    Parameters:
        x (int): First index.
        y (int): Second index.
        z (int): Third index.

    Returns:
        float: The absolute value of the Levi-Civita symbol.
    """
    return (x - y) * (y - z) * (z - x) / 2


# Build the Raman tensor using the absolute values of the Levi-Civita symbol
raman_tensor = np.zeros([3, 3, 3])
for ix in range(3):
    for iy in range(3):
        for iz in range(3):
            raman_tensor[ix, iy, iz] = np.abs(levi_civita(ix - 1, iy - 1, iz - 1))


# Input parameters from command line
eta = float(sys.argv[1])
ortho = False

k = 100  # Lanczos chain length
root = "fc/"
logging.info(root)
atoms = read(root + "replicated_atoms.xyz")
mass = atoms.get_masses()

threshold = 1e-14

# Try loading the dynamical matrix; if not found, build it from the second derivative
# matrix.
try:
    dynmat_sparse = sparse.load_npz(root + "dynmat.npz")
    # Symmetrize matrix for numerical reasons
    dynmat_sparse = (dynmat_sparse + dynmat_sparse.transpose()) / 2
except Exception as e:
    logging.info("Exception: " + str(e))
    try:
        second_sparse = sparse.load_npz(root + "second.npz")
        dynmat_sparse = second2dynmat_sparse(second_sparse=second_sparse, mass=mass)
        dynmat_sparse = (dynmat_sparse + dynmat_sparse.transpose()) / 2
        sparse.save_npz(root + "dynmat.npz", dynmat_sparse)
    except Exception as e:
        logging.info("Exception: " + str(e))
        dynmat = np.load(root + "second.npy")
        dynmat = np.where(np.abs(dynmat) > threshold, dynmat, 0)
        sparsity = 1.0 - np.count_nonzero(dynmat) / dynmat.size
        logging.info("sparsity " + str(sparsity))

        dynmat = dynmat / np.sqrt(
            mass[np.newaxis, :, np.newaxis, np.newaxis, np.newaxis, np.newaxis]
        )
        dynmat = dynmat / np.sqrt(
            mass[np.newaxis, np.newaxis, np.newaxis, np.newaxis, :, np.newaxis]
        )
        n_modes = dynmat.shape[1] * dynmat.shape[2]
        logging.info("nmodes= " + str(n_modes))
        dynmat *= units.mol / (10 * units.J)
        dynmat = dynmat.reshape([n_modes, n_modes])
        dynmat = (dynmat + dynmat.T) / 2
        dynmat_sparse = sparse.csr_matrix(dynmat)
        sparse.save_npz(root + "dynmat.npz", dynmat_sparse)


# Define polarization directions (based on physical review paper)
raman = {}
pol_dir = {}
pol_dir["xy"] = np.array([[0, 1, 0], [1, 0, 0], [0, 0, 0]]) / np.sqrt(2)
pol_dir["xz"] = np.array([[0, 0, 1], [0, 0, 0], [1, 0, 0]]) / np.sqrt(2)
pol_dir["yz"] = np.array([[0, 0, 0], [0, 0, 1], [0, 1, 0]]) / np.sqrt(2)
pol_dir["A"] = np.eye(3) / np.sqrt(3)
pol_dir["1"] = np.array([[1, 0, 0], [0, -1, 0], [0, 0, 0]]) / np.sqrt(3)
pol_dir["2"] = np.array([[1, 0, 0], [0, 0, 0], [0, 0, -1]]) / np.sqrt(3)
pol_dir["3"] = np.array([[0, 0, 0], [0, 1, 0], [0, 0, -1]]) / np.sqrt(3)
pol_dir["4"] = np.array([[0, 1, 0], [-1, 0, 0], [0, 0, 0]]) / np.sqrt(2)
pol_dir["5"] = np.array([[0, 0, 1], [0, 0, 0], [-1, 0, 0]]) / np.sqrt(2)
pol_dir["6"] = np.array([[0, 0, 0], [0, 0, 1], [0, -1, 0]]) / np.sqrt(2)

n_atoms = atoms.positions.shape[0]
polb = np.array([(-1) ** i for i in range(n_atoms)])
m_sqrt = np.sqrt(mass)

# Calculate Raman vectors for each polarization direction
for dire in pol_dir.keys():
    # Compute the contraction between the polarization direction and Raman tensor
    in_out = pol_dir[dire]
    lc = np.einsum("xy,xyz->z", in_out, raman_tensor)
    if np.linalg.norm(lc) < 1e-6:
        continue
    logging.info(str(dire) + " " + str(lc))
    raman_ = np.einsum("i,i,g->ig", polb, 1 / m_sqrt, lc)
    raman_ = raman_.reshape([n_atoms * 3])
    raman[dire] = raman_ / np.linalg.norm(raman_)

# Compute spectrum for a range of frequencies
spectrum = {}
omega_array = np.linspace(0.1, 120, 10000)
spectrum["omega"] = omega_array

result = []
for dire in raman.keys():
    start = time()
    ket = raman[dire]
    result_ = lanczos.spectrum(dynmat_sparse, ket, k, omega_array, eta, True, ortho)
    end = time()
    logging.info("Polarization " + dire + " done in {} s".format(end - start))
    result.append(result_)
result = np.array(result)

# Populate the spectrum with Raman results
for idire, dire in enumerate(raman.keys()):
    spectrum[dire] = {
        "S": result[idire, 0],
        "alpha": result[idire, 1],
        "beta": result[idire, 2],
    }

# Save the spectrum results to file
root_spec = "./spectrum/"
if not os.path.exists(root_spec):
    os.makedirs(root_spec)
if ortho:
    np.save(root_spec + "raman_ortho_k{}_eta{}.npy".format(k, eta), spectrum)
else:
    np.save(root_spec + "raman_k{}_eta{}.npy".format(k, eta), spectrum)
