#!/bin/bash
# This script copies necessary Python scripts to designated directories,
# generates a SLURM job submission file for each configuration, and submits the job.
# Each job performs various calculations (VDSF, Raman, thermal conductivity) using Python.
#
# Set paths
path='../generation_scripts/'
root=$PWD
echo "Using generation scripts at: $path and working directory: $root"

# Loop over concentration (c), system size (N), and instance index (i)
for c in 0.5; do
    for N in 12; do
        for i in 0; do
            eta=0.5
            tempo=1
            partition='g100_usr_dbg'
            
            # Adjust resources based on system size
            if [[ $N -gt 20 ]]; then
                    partition='g100_usr_prod'
                    tempo=24
            fi
            if [[ $N -gt 51 ]]; then
                    partition='g100_usr_bmem'
            fi
            
            # Copy Python scripts to the target directory
            target_dir="${path}/c_${c}/N${N}/${i}"
            cp python_scripts/create_path.py "$target_dir"
            cp python_scripts/path_DSF.py "$target_dir"
            cp python_scripts/raman.py "$target_dir"
            cp python_scripts/TC.py "$target_dir"
            
            # Change to the target directory
            cd "$target_dir" || { echo "Directory $target_dir not found"; exit 1; }
            
            # Create SLURM job submission file
            cat > submit.job <<EOF
#!/bin/bash
#SBATCH --nodes=1                 # number of nodes
##SBATCH --ntasks-per-node=4      # number of tasks per node
##SBATCH --ntasks-per-socket=4    # number of tasks per socket
##SBATCH --cpus-per-task=12       # number of HW threads per task
##SBATCH --mem=230000MB
#SBATCH --time ${tempo}:00:00       # job time limit (HH:MM:SS)
#SBATCH --exclusive
#SBATCH -A Sis25_baroni 
#SBATCH --job-name=silica
#SBATCH -p ${partition}
#SBATCH --output=total.out
#SBATCH --mail-user=afiorent@sissa.it

# Change directory to the submission directory
cd "\${SLURM_SUBMIT_DIR}" 
echo "Submission directory: \${SLURM_SUBMIT_DIR}"
echo "Job start time: \$(date)"

# Setup conda environment
eval "\$(/g100/home/userexternal/afiorent/miniconda3/bin/conda shell.bash hook)"
conda activate nep-kaldo

# Set PYTHONPATH and environment variables
export PYTHONPATH=../../../../../kaldo/:\$PYTHONPATH
export PYTHONPATH=../../../../../hydro_glasses/:\$PYTHONPATH
export OMP_NUM_THREADS=1

echo "Job triggered at: \$(date)"

# Calculations
echo "Calculating VDSF"
python3 create_path.py ${c} ${N}
python3 path_DSF.py ${eta}

echo "Calculating Raman"
python3 raman.py 0.94

echo "Calculating thermal conductivity"
python3 TC.py ${c}

echo "Job finished at: \$(date)"
EOF
            
            # Submit the job and return to the original directory
            sbatch submit.job
            cd "$root" || exit
        done
    done
done
