"""
TC.py
-------
Computes anharmonic properties and thermal conductivity for silicon bulk using hiPhive.
This script loads force constants, computes phonon frequencies at several temperatures,
and then calculates the thermal conductivity using the QHGK method.
"""

import sys
import logging
import pickle
import numpy as np
from kaldo.conductivity import Conductivity
from kaldo.forceconstants import ForceConstants
from kaldo.phonons import Phonons

logging.basicConfig(level=logging.INFO)


def main():
    """
    Main function to compute thermal conductivity.
    Expects one command-line argument to set a parameter 'c'.
    """
    try:
        # Parse command line argument for parameter 'c'
        c = float(sys.argv[1])
    except (IndexError, ValueError):
        sys.exit("Usage: python TC.py <parameter c>")

    if c in (0, 1):
        c = int(c)

    # Load force constants from the 'fc/' folder using hiPhive format
    # (only second order)
    forceconstants = ForceConstants.from_folder(
        "fc/", supercell=[1, 1, 1], format="numpy", only_second=True
    )

    # Define k-point mesh
    k_points = 1  # Number of k points in each direction

    # Loop over temperatures (in Kelvin)
    for temperature in [300, 450, 600, 750, 900]:
        # Configuration for phonon calculations
        phonons_config = {
            "kpts": [k_points, k_points, k_points],
            "is_classic": False,
            "temperature": temperature,
            "folder": "ald",
            "storage": "numpy",
        }
        # Set up phonon object using the force constants
        phonons = Phonons(forceconstants=forceconstants, **phonons_config)
        freq = phonons.frequency

        # Compute thermal conductivity via the QHGK method
        qhgk_cond = Conductivity(phonons=phonons, method="qhgk", storage="numpy")

        # Load the spline data for interpolator
        root_int = "../../../../../bw/interpolator_T{}_c{}.pkl".format(temperature, c)
        logging.info("Spline loaded from: " + root_int)
        with open(root_int, "rb") as f:
            spl = pickle.load(f)

        gamma = spl(freq.flatten()).reshape(freq.shape)
        # Set diffusivity bandwidth to half of gamma
        qhgk_cond.diffusivity_bandwidth = 0.5 * gamma

        # Output results
        conductivity_sum = qhgk_cond.conductivity.sum(axis=0)
        mean_cond = np.mean(np.diag(conductivity_sum))
        print("Conductivity from QHGK (W/m-K): %.3f" % mean_cond)
        print(conductivity_sum)


if __name__ == "__main__":
    main()
